#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <string.h>
#include <stdlib.h>

//#define TESTMODE

HWND hWnd = NULL;

struct funckey {
	unsigned short scancode;	// VK for registering
	char *name;			// name used in config file

	char *prog;			// program to run from config file
	char *descr;			// description

	char *shiftprog;		// program to run from config file when shift is pressed in conjunction
	char *shiftdescr;		// description when shift is pressed in conjunction
};

#define numfuncs (sizeof(funcs)/sizeof(funcs[0]))
struct funckey funcs[] = {
	{ 0xAC, "WWW", NULL, NULL, NULL, NULL },
	{ 0xB4, "Email", NULL, NULL, NULL, NULL },
	{ 0xAA, "Search", NULL, NULL, NULL, NULL },
	{ 0xA6, "BrBack", NULL, NULL, NULL, NULL },
	{ 0xA7, "BrForward", NULL, NULL, NULL, NULL },
	{ 0xA9, "BrStop", NULL, NULL, NULL, NULL },
	{ 0xA8, "BrRefresh", NULL, NULL, NULL, NULL },
	{ 0xAB, "BrHistory", NULL, NULL, NULL, NULL },
	{ 0xB5, "MediaPlayer", NULL, NULL, NULL, NULL },
	{ 0xAD, "Mute", NULL, NULL, NULL, NULL },
	{ 0xB3, "Play", NULL, NULL, NULL, NULL },
	{ 0xB2, "Stop", NULL, NULL, NULL, NULL },
	{ 0xB1, "SkipBack", NULL, NULL, NULL, NULL },
	{ 0xB0, "SkipForward", NULL, NULL, NULL, NULL },
	{ 0xAE, "VolumeDown", NULL, NULL, NULL, NULL },
	{ 0xAF, "VolumeUp", NULL, NULL, NULL, NULL },
	{ 0xB6, "MyComputer", NULL, NULL, NULL, NULL },
	{ 0xB7, "Calculator", NULL, NULL, NULL, NULL },
	{ 0x5F, "Sleep", NULL, NULL, NULL, NULL }
};

#define INIFILE		".\\keyfunc.ini"

LRESULT CALLBACK WProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
	char msg[256] = "", *prog;
	int shift=0;
	STARTUPINFO si;
	PROCESS_INFORMATION pi;
	
	switch (uMsg) {
#ifdef TESTMODE
		case WM_KEYDOWN:
			sprintf(msg, "%x %x", wParam, lParam);
			MessageBox(hWnd, msg, "KeyFunc", MB_OK);
			break;
#endif
		case WM_HOTKEY:
			shift = (wParam & 0x1000) != 0;
			wParam &= 0x0fff;
			if (wParam < 0 || wParam >= numfuncs) break;

			if (shift) prog = funcs[wParam].shiftprog;
			else prog = funcs[wParam].prog;
			
			if (!prog) break;

			ZeroMemory(&si, sizeof(si));
			si.cb = sizeof(STARTUPINFO);
			ZeroMemory(&pi, sizeof(pi));
			if (!CreateProcess(NULL, prog, NULL, NULL, FALSE, 0, NULL, NULL, &si, &pi)) {
				sprintf(msg, "There was a problem starting \"%s\", error code %d.",
						prog, GetLastError());
				MessageBox(hWnd, msg, "KeyFunc", MB_OK|MB_ICONEXCLAMATION);
			}
			break;
		
		case WM_CLOSE:
			PostQuitMessage(0);
			break;
	}

	return DefWindowProc(hWnd, uMsg, wParam, lParam);
}


extern int __cdecl installhook(void);
extern void __cdecl releasehook(void);

int WINAPI WinMain(HINSTANCE hInst, HINSTANCE hPrevInst, LPSTR lpCmdLine, int nCmdShow)
{
	WNDCLASSEX wcx;
	MSG msg;
	BOOL ret;
	int i, hookinstalled = 0;
	char str[1024], *str2;

	hWnd = FindWindow("KeyFunc",NULL);
	if (!strcasecmp(lpCmdLine, "/?") || !strcasecmp(lpCmdLine, "/h")) {
		strcpy(str,
			"KeyFunc by Jonathon Fowler\n"
			"Program to capture the extra function keys on modern 104-key keyboards.\n\n"
			"Parameters:\n"
			"\t/u\tUnload\n"
			"\t/h\tHelp\n\n"
			"Initialization file:\n"
			"\t[Functions]\n"
			"\tName=Program\n"
			"\t[Descriptions]\n"
			"\tName=Description\n\n"
			"\t[ShiftFunctions]\n"
			"\tName=Program\n"
			"\t[ShiftDescriptions]\n"
			"\tName=Description\n\n"
			"\'Name\' may be one of:\n"
		);
		for (i=0; i<numfuncs; i++)
		{
			if (i>0) strcat(str, ", ");
			if (i%15==0) strcat(str, "\n");
			strcat(str, funcs[i].name);
		}
		MessageBox(0,
			str,
			"KeyFunc",
			MB_OK);
		return 0;
	} else if (!strcasecmp(lpCmdLine, "/u")) {
		if (hWnd) {
			SendMessage(hWnd, WM_CLOSE, 0, 0);
			MessageBox(0, "KeyFunc unloaded", "KeyFunc", MB_OK|MB_ICONINFORMATION);
		} else {
			MessageBox(0, "KeyFunc is not loaded", "KeyFunc", MB_OK|MB_ICONINFORMATION);
		}
		return 0;
	} else {
		if (hWnd) {
			MessageBox(hWnd, "Only one instance of KeyFunc may run at a time.", "KeyFunc", MB_OK|MB_ICONINFORMATION);
			return 0;
		}
	}

	wcx.cbSize = sizeof(WNDCLASSEX);
	wcx.style  = 0;
	wcx.lpfnWndProc = WProc;
	wcx.cbClsExtra = 0;
	wcx.cbWndExtra = 0;
	wcx.hInstance = hInst;
	wcx.hIcon = LoadIcon(NULL, IDI_APPLICATION);
	wcx.hCursor = LoadCursor(NULL, IDC_ARROW);
	wcx.hbrBackground = (HBRUSH)COLOR_BTNFACE;
	wcx.lpszMenuName = NULL;
	wcx.lpszClassName = "KeyFunc";
	wcx.hIconSm = LoadIcon(NULL, IDI_APPLICATION);
	if (!RegisterClassEx(&wcx)) {
		MessageBox(0, "Failed registering window class","KeyFunc",MB_OK);
		return -1;
	}

#ifdef TESTMODE
#define WINSTYLE (WS_OVERLAPPED | WS_SYSMENU | WS_VISIBLE)
#else
#define WINSTYLE (WS_OVERLAPPED | WS_SYSMENU)
#endif
	
	hWnd = CreateWindow(
		"KeyFunc",
		"KeyFunc",
		WINSTYLE,
		CW_USEDEFAULT,
		CW_USEDEFAULT,
		CW_USEDEFAULT,
		CW_USEDEFAULT,
		NULL,
		NULL,
		hInst,
		0);
	if (!hWnd) {
		MessageBox(0, "Failed opening window","KeyFunc",MB_OK);
		return -1;
	}

	{
		INT usehook;
		usehook = GetPrivateProfileInt("General","InstallHook",0,INIFILE);
		if (usehook) {
			// install the keyboard hook
			hookinstalled = installhook();
		}
	}

	for (i=0; i<numfuncs; i++) {
		if (GetPrivateProfileString("Functions", funcs[i].name, "", str, sizeof(str), INIFILE) > 0)
			funcs[i].prog = strdup(str);
		if (GetPrivateProfileString("ShiftFunctions", funcs[i].name, "", str, sizeof(str), INIFILE) > 0)
			funcs[i].shiftprog = strdup(str);

		str2 = NULL;
		if (GetPrivateProfileString("Descriptions", funcs[i].name, "", str, sizeof(str), INIFILE) > 0)
			str2 = strdup(str);
		if (str2) funcs[i].descr = str2;
		else funcs[i].descr = funcs[i].name;

		str2 = NULL;
		if (GetPrivateProfileString("ShiftDescriptions", funcs[i].name, "", str, sizeof(str), INIFILE) > 0)
			str2 = strdup(str);
		if (str2) funcs[i].shiftdescr = str2;
		else funcs[i].shiftdescr = funcs[i].name;
		
		if (funcs[i].prog) RegisterHotKey(hWnd, i, 0, funcs[i].scancode);
		if (funcs[i].shiftprog) RegisterHotKey(hWnd, 0x1000|i, MOD_SHIFT, funcs[i].scancode);
	}

	while ((ret = GetMessage(&msg, NULL, 0, 0)) != 0) {
		if (ret == -1) {
			
		} else {
			TranslateMessage(&msg);
			DispatchMessage(&msg);
		}
	}

	if (hookinstalled) {
		// uninstall hook
		releasehook();
	}

	for (i=0; i<numfuncs; i++) {
		if (funcs[i].prog) {
			UnregisterHotKey(hWnd, i);
			free(funcs[i].prog);
		}
		if (funcs[i].shiftprog) {
			UnregisterHotKey(hWnd, 0x1000|i);
			free(funcs[i].shiftprog);
		}
		if (funcs[i].descr != funcs[i].name)
			free(funcs[i].descr);
	}

	return 0;
}


